@@ -26,24 +26,24 @@
 //!
 //! # Examples
 //!
-//! ~~~rust
+//! ```rust
 //! use chainerror::*;
 //! use std::error::Error;
 //! use std::io;
 //! use std::result::Result;
 //!
-//! fn do_some_io() -> Result<(), Box<Error>> {
+//! fn do_some_io() -> Result<(), Box<dyn Error + Send + Sync>> {
 //!     Err(io::Error::from(io::ErrorKind::NotFound))?;
 //!     Ok(())
 //! }
 //!
-//! fn func2() -> Result<(), Box<Error>> {
+//! fn func2() -> Result<(), Box<dyn Error + Send + Sync>> {
 //!     let filename = "foo.txt";
 //!     do_some_io().map_err(mstrerr!("Error reading '{}'", filename))?;
 //!     Ok(())
 //! }
 //!
-//! fn func1() -> Result<(), Box<Error>> {
+//! fn func1() -> Result<(), Box<dyn Error + Send + Sync>> {
 //!     func2().map_err(mstrerr!("func1 error"))?;
 //!     Ok(())
 //! }
@@ -65,21 +65,21 @@
 //! #        unreachable!();
 //! #    }
 //! }
-//! ~~~
+//! ```
 //!
 //!
-//! ~~~rust
+//! ```rust
 //! use chainerror::*;
 //! use std::error::Error;
 //! use std::io;
 //! use std::result::Result;
 //!
-//! fn do_some_io() -> Result<(), Box<Error>> {
+//! fn do_some_io() -> Result<(), Box<dyn Error + Send + Sync>> {
 //!     Err(io::Error::from(io::ErrorKind::NotFound))?;
 //!     Ok(())
 //! }
 //!
-//! fn func3() -> Result<(), Box<Error>> {
+//! fn func3() -> Result<(), Box<dyn Error + Send + Sync>> {
 //!     let filename = "foo.txt";
 //!     do_some_io().map_err(mstrerr!("Error reading '{}'", filename))?;
 //!     Ok(())
@@ -144,8 +144,8 @@
 //!         assert!(e.root_cause().is_some());
 //!
 //!         if let Some(e) = e.root_cause() {
-//!             let ioerror = e.downcast_ref::<io::Error>().unwrap();
-//!             eprintln!("\nThe root cause was: std::io::Error: {:#?}", ioerror);
+//!             let io_error = e.downcast_ref::<io::Error>().unwrap();
+//!             eprintln!("\nThe root cause was: std::io::Error: {:#?}", io_error);
 //!         }
 //!
 //!         #[cfg(not(windows))]
@@ -165,7 +165,30 @@
 //! #        unreachable!();
 //! #    }
 //! }
-//! ~~~
+//! ```
+
+#![deny(
+    warnings,
+    absolute_paths_not_starting_with_crate,
+    deprecated_in_future,
+    keyword_idents,
+    macro_use_extern_crate,
+    missing_debug_implementations,
+    trivial_numeric_casts,
+    unused_extern_crates,
+    unused_import_braces,
+    unused_qualifications,
+    unused_results,
+    unused_labels,
+    unused_lifetimes,
+    unstable_features,
+    unreachable_pub,
+    future_incompatible,
+    missing_copy_implementations,
+    missing_doc_code_examples,
+    rust_2018_idioms,
+    rust_2018_compatibility
+)]
 
 use std::any::TypeId;
 use std::error::Error;
@@ -174,9 +197,9 @@ use std::fmt::{Debug, Display, Formatter, Result};
 /// chains an inner error kind `T` with a causing error
 pub struct ChainError<T> {
     #[cfg(not(feature = "no-fileline"))]
-    occurrence: Option<(u32, &'static str)>,
+    occurrence: Option<&'static str>,
     kind: T,
-    error_cause: Option<Box<dyn Error + 'static>>,
+    error_cause: Option<Box<dyn Error + 'static + Send + Sync>>,
 }
 
 /// convenience type alias
@@ -185,10 +208,11 @@ pub type ChainResult<O, E> = std::result::Result<O, ChainError<E>>;
 impl<T: 'static + Display + Debug> ChainError<T> {
     #[cfg(not(feature = "no-fileline"))]
     /// Use the `cherr!()` or `mstrerr!()` macro instead of calling this directly
+    #[inline]
     pub fn new(
         kind: T,
-        error_cause: Option<Box<dyn Error + 'static>>,
-        occurrence: Option<(u32, &'static str)>,
+        error_cause: Option<Box<dyn Error + 'static + Send + Sync>>,
+        occurrence: Option<&'static str>,
     ) -> Self {
         Self {
             occurrence,
@@ -199,10 +223,11 @@ impl<T: 'static + Display + Debug> ChainError<T> {
 
     #[cfg(feature = "no-fileline")]
     /// Use the `cherr!()` or `mstrerr!()` macro instead of calling this directly
+    #[inline]
     pub fn new(
         kind: T,
-        error_cause: Option<Box<dyn Error + 'static>>,
-        _occurrence: Option<(u32, &'static str)>,
+        error_cause: Option<Box<dyn Error + 'static + Send + Sync>>,
+        _occurrence: Option<&'static str>,
     ) -> Self {
         Self { kind, error_cause }
     }
@@ -216,20 +241,19 @@ impl<T: 'static + Display + Debug> ChainError<T> {
     ///
     /// # Examples
     ///
-    /// ~~~rust
-    /// # use crate::chainerror::*;
-    /// # use std::error::Error;
-    /// # use std::io;
-    /// # use std::result::Result;
-    /// #
-    /// fn do_some_io() -> Result<(), Box<Error>> {
+    /// ```rust
+    /// use chainerror::*;
+    /// use std::error::Error;
+    /// use std::io;
+    ///
+    /// fn do_some_io() -> Result<(), Box<dyn Error + Send + Sync>> {
     ///     Err(io::Error::from(io::ErrorKind::NotFound))?;
     ///     Ok(())
     /// }
     ///
     /// derive_str_cherr!(Func2Error);
     ///
-    /// fn func2() -> Result<(), Box<Error>> {
+    /// fn func2() -> Result<(), Box<dyn Error + Send + Sync>> {
     ///     let filename = "foo.txt";
     ///     do_some_io().map_err(mstrerr!(Func2Error, "Error reading '{}'", filename))?;
     ///     Ok(())
@@ -237,7 +261,7 @@ impl<T: 'static + Display + Debug> ChainError<T> {
     ///
     /// derive_str_cherr!(Func1Error);
     ///
-    /// fn func1() -> Result<(), Box<Error>> {
+    /// fn func1() -> Result<(), Box<dyn Error + Send + Sync>> {
     ///     func2().map_err(mstrerr!(Func1Error, "func1 error"))?;
     ///     Ok(())
     /// }
@@ -258,7 +282,8 @@ impl<T: 'static + Display + Debug> ChainError<T> {
     /// #         unreachable!();
     /// #    }
     /// }
-    /// ~~~
+    /// ```
+    #[inline]
     pub fn find_cause<U: Error + 'static>(&self) -> Option<&U> {
         self.iter().filter_map(Error::downcast_ref::<U>).next()
     }
@@ -269,13 +294,17 @@ impl<T: 'static + Display + Debug> ChainError<T> {
     ///
     /// # Examples
     ///
-    /// ~~~rust,ignore
+    /// ```rust
+    /// # use chainerror::*;
+    /// # derive_str_cherr!(FooError);
+    /// # let err = ChainError::new(String::new(), None, None);
     /// // Instead of writing
     /// err.find_cause::<ChainError<FooError>>();
     ///
     /// // leave out the ChainError<FooError> implementation detail
     /// err.find_chain_cause::<FooError>();
-    /// ~~~
+    /// ```
+    #[inline]
     pub fn find_chain_cause<U: Error + 'static>(&self) -> Option<&ChainError<U>> {
         self.iter()
             .filter_map(Error::downcast_ref::<ChainError<U>>)
@@ -288,7 +317,10 @@ impl<T: 'static + Display + Debug> ChainError<T> {
     ///
     /// # Examples
     ///
-    /// ~~~rust,ignore
+    /// ```rust
+    /// # use chainerror::*;
+    /// # derive_str_cherr!(FooErrorKind);
+    /// # let err = ChainError::new(String::new(), None, None);
     /// // Instead of writing
     /// err.find_cause::<ChainError<FooErrorKind>>();
     /// // and/or
@@ -297,8 +329,9 @@ impl<T: 'static + Display + Debug> ChainError<T> {
     /// err.find_cause::<FooErrorKind>();
     ///
     /// // leave out the ChainError<FooErrorKind> implementation detail
-    /// err.find_chain_or_kind::<FooErrorKind>();
-    /// ~~~
+    /// err.find_kind_or_cause::<FooErrorKind>();
+    /// ```
+    #[inline]
     pub fn find_kind_or_cause<U: Error + 'static>(&self) -> Option<&U> {
         self.iter()
             .filter_map(|e| {
@@ -313,20 +346,19 @@ impl<T: 'static + Display + Debug> ChainError<T> {
     ///
     /// # Examples
     ///
-    /// ~~~rust
-    /// # use crate::chainerror::*;
-    /// # use std::error::Error;
-    /// # use std::io;
-    /// # use std::result::Result;
-    /// #
-    /// fn do_some_io() -> Result<(), Box<Error>> {
+    /// ```rust
+    /// use chainerror::*;
+    /// use std::error::Error;
+    /// use std::io;
+    ///
+    /// fn do_some_io() -> Result<(), Box<dyn Error + Send + Sync>> {
     ///     Err(io::Error::from(io::ErrorKind::NotFound))?;
     ///     Ok(())
     /// }
     ///
     /// derive_str_cherr!(Func2Error);
     ///
-    /// fn func2() -> Result<(), Box<Error>> {
+    /// fn func2() -> Result<(), Box<dyn Error + Send + Sync>> {
     ///     let filename = "foo.txt";
     ///     do_some_io().map_err(mstrerr!(Func2Error, "Error reading '{}'", filename))?;
     ///     Ok(())
@@ -357,7 +389,7 @@ impl<T: 'static + Display + Debug> ChainError<T> {
     /// fn main() {
     ///     if let Err(e) = func1() {
     ///         match e.kind() {
-    ///             Func1ErrorKind::Func2 => {},
+    ///             Func1ErrorKind::Func2 => {}
     ///             Func1ErrorKind::IO(filename) => panic!(),
     ///         }
     ///     }
@@ -365,7 +397,8 @@ impl<T: 'static + Display + Debug> ChainError<T> {
     /// #         unreachable!();
     /// #    }
     /// }
-    /// ~~~
+    /// ```
+    #[inline]
     pub fn kind(&self) -> &T {
         &self.kind
     }
@@ -375,6 +408,7 @@ impl<T: 'static + Display + Debug> ChainError<T> {
     /// # Example
     ///
     ///
+    #[inline]
     pub fn iter(&self) -> impl Iterator<Item = &(dyn Error + 'static)> {
         ErrorIter {
             current: Some(self),
@@ -389,6 +423,7 @@ struct ErrorIter<'a> {
 impl<'a> Iterator for ErrorIter<'a> {
     type Item = &'a (dyn Error + 'static);
 
+    #[inline]
     fn next(&mut self) -> Option<Self::Item> {
         let current = self.current;
         self.current = self.current.and_then(Error::source);
@@ -399,6 +434,7 @@ impl<'a> Iterator for ErrorIter<'a> {
 impl<T: 'static + Display + Debug> std::ops::Deref for ChainError<T> {
     type Target = T;
 
+    #[inline]
     fn deref(&self) -> &Self::Target {
         &self.kind
     }
@@ -412,17 +448,24 @@ pub trait ChainErrorDown {
     fn downcast_chain_ref<T: 'static + Display + Debug>(&self) -> Option<&ChainError<T>>;
     /// Downcast to a mutable reference of `ChainError<T>`
     fn downcast_chain_mut<T: 'static + Display + Debug>(&mut self) -> Option<&mut ChainError<T>>;
+    /// Downcast to T of `ChainError<T>`
+    fn downcast_inner_ref<T: 'static + Error>(&self) -> Option<&T>;
+    /// Downcast to T mutable reference of `ChainError<T>`
+    fn downcast_inner_mut<T: 'static + Error>(&mut self) -> Option<&mut T>;
 }
 
 impl<U: 'static + Display + Debug> ChainErrorDown for ChainError<U> {
+    #[inline]
     fn is_chain<T: 'static + Display + Debug>(&self) -> bool {
         TypeId::of::<T>() == TypeId::of::<U>()
     }
 
+    #[inline]
     fn downcast_chain_ref<T: 'static + Display + Debug>(&self) -> Option<&ChainError<T>> {
         if self.is_chain::<T>() {
             #[allow(clippy::cast_ptr_alignment)]
             unsafe {
+                #[allow(trivial_casts)]
                 Some(&*(self as *const dyn Error as *const &ChainError<T>))
             }
         } else {
@@ -430,80 +473,174 @@ impl<U: 'static + Display + Debug> ChainErrorDown for ChainError<U> {
         }
     }
 
+    #[inline]
     fn downcast_chain_mut<T: 'static + Display + Debug>(&mut self) -> Option<&mut ChainError<T>> {
         if self.is_chain::<T>() {
             #[allow(clippy::cast_ptr_alignment)]
             unsafe {
+                #[allow(trivial_casts)]
                 Some(&mut *(self as *mut dyn Error as *mut &mut ChainError<T>))
             }
         } else {
             None
         }
     }
+    #[inline]
+    fn downcast_inner_ref<T: 'static + Error>(&self) -> Option<&T> {
+        if self.is_chain::<T>() {
+            #[allow(clippy::cast_ptr_alignment)]
+            unsafe {
+                #[allow(trivial_casts)]
+                Some(&(*(self as *const dyn Error as *const &ChainError<T>)).kind)
+            }
+        } else {
+            None
+        }
+    }
+
+    #[inline]
+    fn downcast_inner_mut<T: 'static + Error>(&mut self) -> Option<&mut T> {
+        if self.is_chain::<T>() {
+            #[allow(clippy::cast_ptr_alignment)]
+            unsafe {
+                #[allow(trivial_casts)]
+                Some(&mut (*(self as *mut dyn Error as *mut &mut ChainError<T>)).kind)
+            }
+        } else {
+            None
+        }
+    }
 }
 
 impl ChainErrorDown for dyn Error + 'static {
+    #[inline]
     fn is_chain<T: 'static + Display + Debug>(&self) -> bool {
         self.is::<ChainError<T>>()
     }
 
+    #[inline]
     fn downcast_chain_ref<T: 'static + Display + Debug>(&self) -> Option<&ChainError<T>> {
         self.downcast_ref::<ChainError<T>>()
     }
 
+    #[inline]
     fn downcast_chain_mut<T: 'static + Display + Debug>(&mut self) -> Option<&mut ChainError<T>> {
         self.downcast_mut::<ChainError<T>>()
     }
+
+    #[inline]
+    fn downcast_inner_ref<T: 'static + Error>(&self) -> Option<&T> {
+        self.downcast_ref::<T>()
+            .or_else(|| self.downcast_ref::<ChainError<T>>().map(|e| e.kind()))
+    }
+
+    #[inline]
+    fn downcast_inner_mut<T: 'static + Error>(&mut self) -> Option<&mut T> {
+        if self.is::<T>() {
+            return self.downcast_mut::<T>();
+        }
+
+        self.downcast_mut::<ChainError<T>>()
+            .and_then(|e| e.downcast_inner_mut::<T>())
+    }
 }
 
 impl ChainErrorDown for dyn Error + 'static + Send {
+    #[inline]
     fn is_chain<T: 'static + Display + Debug>(&self) -> bool {
         self.is::<ChainError<T>>()
     }
 
+    #[inline]
     fn downcast_chain_ref<T: 'static + Display + Debug>(&self) -> Option<&ChainError<T>> {
         self.downcast_ref::<ChainError<T>>()
     }
 
+    #[inline]
     fn downcast_chain_mut<T: 'static + Display + Debug>(&mut self) -> Option<&mut ChainError<T>> {
         self.downcast_mut::<ChainError<T>>()
     }
+
+    #[inline]
+    fn downcast_inner_ref<T: 'static + Error>(&self) -> Option<&T> {
+        self.downcast_ref::<T>()
+            .or_else(|| self.downcast_ref::<ChainError<T>>().map(|e| e.kind()))
+    }
+
+    #[inline]
+    fn downcast_inner_mut<T: 'static + Error>(&mut self) -> Option<&mut T> {
+        if self.is::<T>() {
+            return self.downcast_mut::<T>();
+        }
+
+        self.downcast_mut::<ChainError<T>>()
+            .and_then(|e| e.downcast_inner_mut::<T>())
+    }
 }
 
 impl ChainErrorDown for dyn Error + 'static + Send + Sync {
+    #[inline]
     fn is_chain<T: 'static + Display + Debug>(&self) -> bool {
         self.is::<ChainError<T>>()
     }
 
+    #[inline]
     fn downcast_chain_ref<T: 'static + Display + Debug>(&self) -> Option<&ChainError<T>> {
         self.downcast_ref::<ChainError<T>>()
     }
 
+    #[inline]
     fn downcast_chain_mut<T: 'static + Display + Debug>(&mut self) -> Option<&mut ChainError<T>> {
         self.downcast_mut::<ChainError<T>>()
     }
+
+    #[inline]
+    fn downcast_inner_ref<T: 'static + Error>(&self) -> Option<&T> {
+        self.downcast_ref::<T>()
+            .or_else(|| self.downcast_ref::<ChainError<T>>().map(|e| e.kind()))
+    }
+
+    #[inline]
+    fn downcast_inner_mut<T: 'static + Error>(&mut self) -> Option<&mut T> {
+        if self.is::<T>() {
+            return self.downcast_mut::<T>();
+        }
+
+        self.downcast_mut::<ChainError<T>>()
+            .and_then(|e| e.downcast_inner_mut::<T>())
+    }
 }
 
 impl<T: 'static + Display + Debug> Error for ChainError<T> {
+    #[inline]
     fn source(&self) -> Option<&(dyn Error + 'static)> {
-        self.error_cause.as_ref().map(|e| e.as_ref())
+        self.error_cause
+            .as_ref()
+            .map(|e| e.as_ref() as &(dyn Error + 'static))
     }
 }
 
 impl<T: 'static + Display + Debug> Error for &ChainError<T> {
+    #[inline]
     fn source(&self) -> Option<&(dyn Error + 'static)> {
-        self.error_cause.as_ref().map(|e| e.as_ref())
+        self.error_cause
+            .as_ref()
+            .map(|e| e.as_ref() as &(dyn Error + 'static))
     }
 }
 
 impl<T: 'static + Display + Debug> Error for &mut ChainError<T> {
+    #[inline]
     fn source(&self) -> Option<&(dyn Error + 'static)> {
-        self.error_cause.as_ref().map(|e| e.as_ref())
+        self.error_cause
+            .as_ref()
+            .map(|e| e.as_ref() as &(dyn Error + 'static))
     }
 }
 
 impl<T: 'static + Display + Debug> Display for ChainError<T> {
-    fn fmt(&self, f: &mut Formatter) -> Result {
+    #[inline]
+    fn fmt(&self, f: &mut Formatter<'_>) -> Result {
         write!(f, "{}", self.kind)?;
 
         #[cfg(feature = "display-cause")]
@@ -518,11 +655,12 @@ impl<T: 'static + Display + Debug> Display for ChainError<T> {
 }
 
 impl<T: 'static + Display + Debug> Debug for ChainError<T> {
-    fn fmt(&self, f: &mut Formatter) -> Result {
+    #[inline]
+    fn fmt(&self, f: &mut Formatter<'_>) -> Result {
         #[cfg(not(feature = "no-fileline"))]
         {
-            if let Some(o) = self.occurrence {
-                write!(f, "{}:{}: ", o.1, o.0)?;
+            if let Some(ref o) = self.occurrence {
+                Display::fmt(o, f)?;
             }
         }
 
@@ -543,19 +681,24 @@ impl<T: 'static + Display + Debug> Debug for ChainError<T> {
     }
 }
 
+/// `ChainErrorFrom<T>` is similar to `From<T>`
 pub trait ChainErrorFrom<T>: Sized {
-    fn chain_error_from(_: T, line_filename: Option<(u32, &'static str)>) -> ChainError<Self>;
+    /// similar to From<T>::from()
+    fn chain_error_from(from: T, line_filename: Option<&'static str>) -> ChainError<Self>;
 }
 
+/// `IntoChainError<T>` is similar to `Into<T>`
 pub trait IntoChainError<T>: Sized {
-    fn into_chain_error(self, line_filename: Option<(u32, &'static str)>) -> ChainError<T>;
+    /// similar to Into<T>::into()
+    fn into_chain_error(self, line_filename: Option<&'static str>) -> ChainError<T>;
 }
 
 impl<T, U> IntoChainError<U> for T
 where
     U: ChainErrorFrom<T>,
 {
-    fn into_chain_error(self, line_filename: Option<(u32, &'static str)>) -> ChainError<U> {
+    #[inline]
+    fn into_chain_error(self, line_filename: Option<&'static str>) -> ChainError<U> {
         U::chain_error_from(self, line_filename)
     }
 }
@@ -565,24 +708,37 @@ where
     T: Into<U>,
     U: 'static + Display + Debug,
 {
-    fn chain_error_from(t: T, line_filename: Option<(u32, &'static str)>) -> ChainError<Self> {
+    #[inline]
+    fn chain_error_from(t: T, line_filename: Option<&'static str>) -> ChainError<Self> {
         let e: U = t.into();
-        ChainError::<_>::new(e, None, line_filename)
+        ChainError::new(e, None, line_filename)
     }
 }
 
+/*
+impl<T, U> ChainErrorFrom<T> for U
+    where
+        T: 'static + Error + Into<Box<T>> + Clone,
+        U: 'static + Display + Debug + From<T>,
+{
+    #[inline]
+    fn chain_error_from(t: T, line_filename: Option<&'static str>) -> ChainError<Self> {
+        ChainError::new(U::from(t.clone()), Some(Box::from(t)), line_filename)
+    }
+}
+*/
+
+/// map into `ChainError<T>` with `T::from(err)`
+///
+/// adds `line!()` and `file!()` information
 #[macro_export]
 macro_rules! minto_cherr {
-    ( ) => {
-        |e| e.into_chain_error(Some((line!(), file!())))
-    };
-}
-
-#[macro_export]
-macro_rules! into_cherr {
-    ( $t:expr ) => {
-        $t.into_chain_error(Some((line!(), file!())))
-    };
+    ( $k:ident ) => (
+        |e| $crate::cherr!(e, $k::from(&e))
+    );
+    ( $enum:ident $(:: $enum_path:ident)* ) => (
+        |e| $crate::cherr!(e, $enum $(:: $enum_path)*::from(&e))
+    );
 }
 
 /// Creates a new `ChainError<T>`
@@ -590,15 +746,13 @@ macro_rules! into_cherr {
 /// # Examples
 ///
 /// Create a new ChainError<FooError>, where `FooError` must implement `Display` and `Debug`.
-/// ~~~rust
+/// ```rust
 /// # use chainerror::*;
-/// #
 /// # #[derive(Debug)]
 /// enum FooError {
 ///     Bar,
 ///     Baz(&'static str),
 /// }
-/// #
 /// # impl ::std::fmt::Display for FooError {
 /// #     fn fmt(&self, f: &mut ::std::fmt::Formatter) -> ::std::fmt::Result {
 /// #         match self {
@@ -620,28 +774,25 @@ macro_rules! into_cherr {
 ///     }
 ///     Ok(())
 /// }
-/// #
 /// # pub fn main() {
 /// #     match func().unwrap_err().kind() {
-/// #         FooError::Baz(s) if s == &"Error" => {},
+/// #         FooError::Baz(s) if s == &"Error" => {}
 /// #         _ => panic!(),
 /// #     }
 /// # }
-/// ~~~
+/// ```
 ///
 /// Additionally an error cause can be added.
 ///
-/// ~~~rust
+/// ```rust
 /// # use chainerror::*;
 /// # use std::io;
 /// # use std::error::Error;
-/// #
 /// # #[derive(Debug)]
 /// # enum FooError {
 /// #     Bar,
 /// #     Baz(&'static str),
 /// # }
-/// #
 /// # impl ::std::fmt::Display for FooError {
 /// #     fn fmt(&self, f: &mut ::std::fmt::Formatter) -> ::std::fmt::Result {
 /// #         match self {
@@ -650,8 +801,7 @@ macro_rules! into_cherr {
 /// #         }
 /// #     }
 /// # }
-/// #
-/// fn do_some_stuff() -> Result<(), Box<Error>> {
+/// fn do_some_stuff() -> Result<(), Box<dyn Error + Send + Sync>> {
 ///     Err(io::Error::from(io::ErrorKind::NotFound))?;
 ///     Ok(())
 /// }
@@ -662,59 +812,55 @@ macro_rules! into_cherr {
 ///     )?;
 ///     Ok(())
 /// }
-/// #
 /// # pub fn main() {
 /// #     match func().unwrap_err().kind() {
-/// #         FooError::Baz(s) if s == &"Error" => {},
+/// #         FooError::Baz(s) if s == &"Error" => {}
 /// #         _ => panic!(),
 /// #     }
 /// # }
-/// ~~~
+/// ```
 #[macro_export]
 macro_rules! cherr {
     ( $k:expr ) => ({
-        ChainError::<_>::new($k, None, Some((line!(), file!())))
+        $crate::ChainError::new($k, None, Some(concat!(file!(), ":", line!(), ": ")))
     });
     ( None, $k:expr ) => ({
-        ChainError::<_>::new($k, None, Some((line!(), file!())))
+        $crate::ChainError::new($k, None, Some(concat!(file!(), ":", line!(), ": ")))
     });
     ( None, $fmt:expr, $($arg:tt)+ ) => ({
-        cherr!(None, format!($fmt, $($arg)+ ))
+        $crate::cherr!(None, format!($fmt, $($arg)+ ))
     });
     ( None, $fmt:expr, $($arg:tt)+ ) => ({
-        cherr!(None, format!($fmt, $($arg)+ ))
+        $crate::cherr!(None, format!($fmt, $($arg)+ ))
     });
     ( $e:path, $k:expr ) => ({
-        ChainError::<_>::new($k, Some(Box::from($e)), Some((line!(), file!())))
+        $crate::ChainError::new($k, Some(Box::from($e)), Some(concat!(file!(), ":", line!(), ": ")))
     });
     ( $e:path, $fmt:expr, $($arg:tt)+ ) => ({
-        cherr!($e, format!($fmt, $($arg)+ ))
+        $crate::cherr!($e, format!($fmt, $($arg)+ ))
     });
-
 }
 
 /// Convenience macro for `|e| cherr!(e, format!(…))`
 ///
 /// # Examples
 ///
-/// ~~~rust
+/// ```rust
 /// # use crate::chainerror::*;
 /// # use std::error::Error;
 /// # use std::io;
 /// # use std::result::Result;
-/// #
-/// # fn do_some_io() -> Result<(), Box<Error>> {
+/// # fn do_some_io() -> Result<(), Box<dyn Error + Send + Sync>> {
 /// #     Err(io::Error::from(io::ErrorKind::NotFound))?;
 /// #     Ok(())
 /// # }
-/// #
-/// fn func2() -> Result<(), Box<Error>> {
+/// fn func2() -> Result<(), Box<dyn Error + Send + Sync>> {
 ///     let filename = "foo.txt";
 ///     do_some_io().map_err(mstrerr!("Error reading '{}'", filename))?;
 ///     Ok(())
 /// }
 ///
-/// fn func1() -> Result<(), Box<Error>> {
+/// fn func1() -> Result<(), Box<dyn Error + Send + Sync>> {
 ///     func2().map_err(mstrerr!("func1 error"))?;
 ///     Ok(())
 /// }
@@ -724,9 +870,9 @@ macro_rules! cherr {
 /// #         #[cfg(not(windows))]
 /// #         assert_eq!(
 /// #             format!("\n{:?}\n", e), r#"
-/// # src/lib.rs:20: func1 error
+/// # src/lib.rs:18: func1 error
 /// # Caused by:
-/// # src/lib.rs:15: Error reading 'foo.txt'
+/// # src/lib.rs:13: Error reading 'foo.txt'
 /// # Caused by:
 /// # Kind(NotFound)
 /// # "#
@@ -735,25 +881,23 @@ macro_rules! cherr {
 /// #         unreachable!();
 /// #     }
 /// # }
-/// ~~~
+/// ```
 ///
 /// `mstrerr!()` can also be used to map a new `ChainError<T>`, where T was defined with
 /// `derive_str_cherr!(T)`
 ///
-/// ~~~rust
+/// ```rust
 /// # use crate::chainerror::*;
 /// # use std::error::Error;
 /// # use std::io;
 /// # use std::result::Result;
-/// #
-/// # fn do_some_io() -> Result<(), Box<Error>> {
+/// # fn do_some_io() -> Result<(), Box<dyn Error + Send + Sync>> {
 /// #     Err(io::Error::from(io::ErrorKind::NotFound))?;
 /// #     Ok(())
 /// # }
-/// #
 /// derive_str_cherr!(Func2Error);
 ///
-/// fn func2() -> Result<(), Box<Error>> {
+/// fn func2() -> Result<(), Box<dyn Error + Send + Sync>> {
 ///     let filename = "foo.txt";
 ///     do_some_io().map_err(mstrerr!(Func2Error, "Error reading '{}'", filename))?;
 ///     Ok(())
@@ -761,11 +905,10 @@ macro_rules! cherr {
 ///
 /// derive_str_cherr!(Func1Error);
 ///
-/// fn func1() -> Result<(), Box<Error>> {
+/// fn func1() -> Result<(), Box<dyn Error + Send + Sync>> {
 ///     func2().map_err(mstrerr!(Func1Error, "func1 error"))?;
 ///     Ok(())
 /// }
-/// #
 /// # fn main() {
 /// #     if let Err(e) = func1() {
 /// #         if let Some(f1err) = e.downcast_chain_ref::<Func1Error>() {
@@ -778,26 +921,26 @@ macro_rules! cherr {
 /// #         unreachable!();
 /// #     }
 /// # }
-/// ~~~
+/// ```
 #[macro_export]
 macro_rules! mstrerr {
     ( $t:path, $msg:expr ) => ({
-        |e| cherr!(e, $t ($msg.to_string()))
+        |e| $crate::cherr!(e, $t ($msg.to_string()))
     });
     ( $t:path, $msg:expr, ) => ({
-        |e| cherr!(e, $t ($msg.to_string()))
+        |e| $crate::cherr!(e, $t ($msg.to_string()))
     });
     ( $t:path, $fmt:expr, $($arg:tt)+ ) => ({
-        |e| cherr!(e, $t (format!($fmt, $($arg)+ )))
+        |e| $crate::cherr!(e, $t (format!($fmt, $($arg)+ )))
     });
     ($msg:expr) => ({
-        |e| cherr!(e, $msg.to_string())
+        |e| $crate::cherr!(e, $msg.to_string())
     });
     ($msg:expr, ) => ({
-        |e| cherr!(e, $msg.to_string())
+        |e| $crate::cherr!(e, $msg.to_string())
     });
     ($fmt:expr, $($arg:tt)+) => ({
-        |e| cherr!(e, format!($fmt, $($arg)+ ))
+        |e| $crate::cherr!(e, format!($fmt, $($arg)+ ))
     });
 }
 
@@ -805,11 +948,10 @@ macro_rules! mstrerr {
 ///
 /// # Examples
 ///
-/// ~~~rust
+/// ```rust
 /// # use crate::chainerror::*;
 /// # use std::error::Error;
 /// # use std::result::Result;
-/// #
 /// derive_str_cherr!(Func2Error);
 ///
 /// fn func2() -> ChainResult<(), Func2Error> {
@@ -819,11 +961,10 @@ macro_rules! mstrerr {
 ///
 /// derive_str_cherr!(Func1Error);
 ///
-/// fn func1() -> Result<(), Box<Error>> {
+/// fn func1() -> Result<(), Box<dyn Error>> {
 ///     func2().map_err(mstrerr!(Func1Error, "func1 error"))?;
 ///     Ok(())
 /// }
-/// #
 /// # fn main() {
 /// #     if let Err(e) = func1() {
 /// #         if let Some(f1err) = e.downcast_chain_ref::<Func1Error>() {
@@ -836,26 +977,26 @@ macro_rules! mstrerr {
 /// #         unreachable!();
 /// #     }
 /// # }
-/// ~~~
+/// ```
 #[macro_export]
 macro_rules! strerr {
     ( $t:path, $msg:expr ) => ({
-        cherr!($t ($msg.to_string()))
+        $crate::cherr!($t ($msg.to_string()))
     });
     ( $t:path, $msg:expr, ) => ({
-        cherr!($t ($msg.to_string()))
+        $crate::cherr!($t ($msg.to_string()))
     });
     ( $t:path, $fmt:expr, $($arg:tt)+ ) => ({
-        cherr!($t (format!($fmt, $($arg)+ )))
+        $crate::cherr!($t (format!($fmt, $($arg)+ )))
     });
     ($msg:expr) => ({
-        cherr!($msg.to_string())
+        $crate::cherr!($msg.to_string())
     });
     ($msg:expr, ) => ({
-        cherr!($msg.to_string())
+        $crate::cherr!($msg.to_string())
     });
     ($fmt:expr, $($arg:tt)+) => ({
-        cherr!(format!($fmt, $($arg)+ ))
+        $crate::cherr!(format!($fmt, $($arg)+ ))
     });
 }
 
@@ -863,17 +1004,15 @@ macro_rules! strerr {
 ///
 /// # Examples
 ///
-/// ~~~rust
+/// ```rust
 /// # use crate::chainerror::*;
 /// # use std::error::Error;
 /// # use std::io;
 /// # use std::result::Result;
-/// #
-/// # fn do_some_io() -> Result<(), Box<Error>> {
+/// # fn do_some_io() -> Result<(), Box<dyn Error + Send + Sync>> {
 /// #     Err(io::Error::from(io::ErrorKind::NotFound))?;
 /// #     Ok(())
 /// # }
-/// #
 /// derive_str_cherr!(Func2Error);
 ///
 /// fn func2() -> ChainResult<(), Func2Error> {
@@ -884,11 +1023,10 @@ macro_rules! strerr {
 ///
 /// derive_str_cherr!(Func1Error);
 ///
-/// fn func1() -> Result<(), Box<Error>> {
+/// fn func1() -> Result<(), Box<dyn Error>> {
 ///     func2().map_err(mstrerr!(Func1Error, "func1 error"))?;
 ///     Ok(())
 /// }
-/// #
 /// # fn main() {
 /// #     if let Err(e) = func1() {
 /// #         if let Some(f1err) = e.downcast_chain_ref::<Func1Error>() {
@@ -901,10 +1039,11 @@ macro_rules! strerr {
 /// #         unreachable!();
 /// #     }
 /// # }
-/// ~~~
+/// ```
 #[macro_export]
 macro_rules! derive_str_cherr {
     ($e:ident) => {
+        #[derive(Clone)]
         pub struct $e(pub String);
         impl ::std::fmt::Display for $e {
             fn fmt(&self, f: &mut ::std::fmt::Formatter) -> ::std::fmt::Result {
@@ -919,3 +1058,134 @@ macro_rules! derive_str_cherr {
         impl ::std::error::Error for $e {}
     };
 }
+
+/// Derive an Error for an ErrorKind, which wraps a `ChainError` and implements a `kind()` method
+///
+/// It basically hides `ChainError` to the outside and only exposes the `kind()`
+/// method.
+///
+/// Error::kind() returns the ErrorKind
+/// Error::source() returns the parent error
+///
+/// # Examples
+///
+/// ```rust
+/// use std::io;
+/// use chainerror::*;
+///
+/// fn do_some_io(_f: &str) -> std::result::Result<(), io::Error> {
+///     return Err(io::Error::from(io::ErrorKind::NotFound));
+/// }
+///
+/// #[derive(Debug, Clone)]
+/// pub enum ErrorKind {
+///     IO(String),
+///     FatalError(String),
+///     Unknown,
+/// }
+///
+/// derive_err_kind!(Error, ErrorKind);
+///
+/// impl std::fmt::Display for ErrorKind {
+///     fn fmt(&self, f: &mut ::std::fmt::Formatter) -> std::fmt::Result {
+///         match self {
+///             ErrorKind::FatalError(e) => write!(f, "fatal error {}", e),
+///             ErrorKind::Unknown => write!(f, "unknown error"),
+///             ErrorKind::IO(filename) => write!(f, "Error reading '{}'", filename),
+///         }
+///     }
+/// }
+///
+/// impl ErrorKind {
+///     fn from_io_error(e: &io::Error, f: String) -> Self {
+///         match e.kind() {
+///             io::ErrorKind::BrokenPipe => panic!("Should not happen"),
+///             io::ErrorKind::ConnectionReset => {
+///                 ErrorKind::FatalError(format!("While reading `{}`: {}", f, e))
+///             }
+///             _ => ErrorKind::IO(f),
+///         }
+///     }
+/// }
+///
+/// impl From<&io::Error> for ErrorKind {
+///     fn from(e: &io::Error) -> Self {
+///         ErrorKind::IO(format!("{}", e))
+///     }
+/// }
+///
+/// pub fn func1() -> std::result::Result<(), Error> {
+///     let filename = "bar.txt";
+///
+///     do_some_io(filename)
+///         .map_err(|e| cherr!(e, ErrorKind::from_io_error(&e, filename.into())))?;
+///     do_some_io(filename).map_err(|e| cherr!(e, ErrorKind::IO(filename.into())))?;
+///     do_some_io(filename).map_err(|e| cherr!(e, ErrorKind::from(&e)))?;
+///     do_some_io(filename).map_err(minto_cherr!(ErrorKind))?;
+///     Ok(())
+/// }
+/// ```
+#[macro_export]
+macro_rules! derive_err_kind {
+    ($e:ident, $k:ident) => {
+        pub struct $e($crate::ChainError<$k>);
+
+        impl $e {
+            pub fn kind(&self) -> &$k {
+                self.0.kind()
+            }
+        }
+
+        impl From<$k> for $e {
+            fn from(e: $k) -> Self {
+                $e($crate::ChainError::new(e, None, None))
+            }
+        }
+
+        impl From<ChainError<$k>> for $e {
+            fn from(e: $crate::ChainError<$k>) -> Self {
+                $e(e)
+            }
+        }
+
+        impl From<&$e> for $k
+        where
+            $k: Clone,
+        {
+            fn from(e: &$e) -> Self {
+                e.kind().clone()
+            }
+        }
+
+        impl $crate::ChainErrorFrom<$e> for $k
+        where
+            $k: Clone,
+        {
+            #[inline]
+            fn chain_error_from(
+                t: $e,
+                line_filename: Option<&'static str>,
+            ) -> $crate::ChainError<$k> {
+                $crate::ChainError::new((*t.kind()).clone(), Some(Box::from(t)), line_filename)
+            }
+        }
+
+        impl std::error::Error for $e {
+            fn source(&self) -> Option<&(dyn std::error::Error + 'static)> {
+                self.0.source()
+            }
+        }
+
+        impl std::fmt::Display for $e {
+            fn fmt(&self, f: &mut std::fmt::Formatter) -> std::fmt::Result {
+                std::fmt::Display::fmt(&self.0, f)
+            }
+        }
+
+        impl std::fmt::Debug for $e {
+            fn fmt(&self, f: &mut std::fmt::Formatter) -> std::fmt::Result {
+                std::fmt::Debug::fmt(&self.0, f)
+            }
+        }
+    };
+}
